import logging
import os
import boto3  # type: ignore
from botocore.exceptions import ClientError  # type: ignore


def upload_dir_to_s3(local_directory: str, s3_bucket: str, s3_bucket_dir: str,
                     extra_args=None) -> None:

    client = boto3.client('s3')
    for root, dirs, files in os.walk(local_directory):

        for filename in files:
            local_path = os.path.join(root, filename)

            # construct the full path
            relative_path = os.path.relpath(local_path, local_directory)
            s3_path = os.path.join(s3_bucket_dir, relative_path)

            logging.info(f"Searching {s3_path} in {s3_bucket}")
            try:
                client.head_object(Bucket=s3_bucket, Key=s3_path)
                logging.warning("Path found on S3! Skipping {s3_path}")
            except ClientError:  # Exception abuse
                logging.info(f"Uploading {s3_path}")
                client.upload_file(local_path, s3_bucket, s3_path,
                                   ExtraArgs=extra_args)



import pickle
from .model import Model


class SklearnModel(Model):
    def __init__(self, config: dict, outdir: str = None):
        super().__init__(outdir=outdir)
        self.config = config
        model_type = config["classifier_type"]
        model_class = self.dynamically_import_class(model_type)
        self.model = model_class()  # type: ignore

    def fit(self, train_data, test_data):
        return self.model.fit(train_data, test_data)

    def predict_proba(self, predict_data):
        return self.model.predict_proba(predict_data)

    def load(self, path: str) -> tuple(): # type: ignore
        with open(path, "rb") as mf1:
            model_object = pickle.load(mf1)
        return model_object  # , custom_model_object



import json
import os
import click
from grape import Graph # type: ignore
import numpy as np  # type: ignore

from neat_ml.link_prediction.sklearn_model import SklearnModel
from neat_ml.link_prediction.mlp_model import MLPModel

from tqdm import tqdm  # type: ignore

from neat_ml.graph_embedding.graph_embedding import make_node_embeddings
from neat_ml.pre_run_checks.pre_run_checks import pre_run_checks
from neat_ml.update_yaml.update_yaml import do_update_yaml
from neat_ml.upload.upload import upload_dir_to_s3
from neat_ml.visualization.visualization import make_tsne
from neat_ml.run_classifier.run_classifier import predict_links
from neat_ml.yaml_helper.yaml_helper import YamlHelper


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--config",
    required=True,
    default="config.yaml",
    type=click.Path(exists=True),
)
def run(config: str) -> None:
    """Run a NEAT pipeline using the given YAML file [neat.yaml]
    \f

    Args:
        config: Specify the YAML file containing instructions of what ML tasks to perform

    Returns:
        None.

    """

    yhelp = YamlHelper(config)

    # pre run checks for failing early
    if not pre_run_checks(yhelp=yhelp, check_s3_credentials=yhelp.do_upload()):
        raise RuntimeError("Failed pre_run_check")

    # generate embeddings if config has 'embeddings' block
    if yhelp.do_embeddings() and not os.path.exists(yhelp.embedding_outfile()):
        node_embedding_args = yhelp.make_node_embeddings_args()
        make_node_embeddings(**node_embedding_args)

    if yhelp.do_tsne() and not os.path.exists(yhelp.tsne_outfile()):
        graph: Graph = yhelp.load_graph()
        tsne_kwargs = yhelp.make_tsne_args(graph)
        make_tsne(**tsne_kwargs)

    if yhelp.do_classifier():
        for classifier in tqdm(yhelp.classifiers()):

            # Check if classifier already exists
            if os.path.exists(yhelp.classifier_outfile(classifier)):
                classifier_id = classifier["classifier_id"]
                print(f"Found existing classifier: {classifier_id}")
                continue

            model: object = None
            if classifier["classifier_name"].lower() == "neural network":
                model = MLPModel(classifier, outdir=yhelp.outdir())
            elif classifier["classifier_name"].lower() in [
                "decision tree",
                "logistic regression",
                "random forest",
            ]:
                model = SklearnModel(classifier, outdir=yhelp.outdir())
            else:
                raise NotImplementedError(f"{model} isn't implemented yet")

            model.compile()

            train_data, validation_data = model.make_train_valid_data(
                embedding_file=yhelp.embedding_outfile(),
                training_graph_args=yhelp.main_graph_args(),
                validation_args=yhelp.val_graph_args(),
                training_args=yhelp.train_graph_args(),
                edge_method=yhelp.get_edge_embedding_method(classifier),
            )
            history_obj = model.fit(*train_data)
            if type(model) == SklearnModel:
                predicted_labels = model.predict(validation_data[0])
            else:
                predicted_labels = np.concatenate(
                    np.around(model.predict(validation_data[0]), decimals=0)
                )
            actual_labels = validation_data[1]
            correct_matches = sum(list(predicted_labels == actual_labels))
            total_data_points = len(validation_data[0])
            correct_label_match = (correct_matches / total_data_points) * 100

            print(f"Correct label match in validation: {correct_label_match}")

            if yhelp.classifier_history_file_name(classifier):
                with open(yhelp.classifier_history_file_name(classifier), "w") as f:  # type: ignore
                    json.dump(history_obj.history, f)

            model.save()

    if yhelp.do_apply_classifier():
        # take graph, classifier, biolink node types and cutoff
        for clsfr_id in yhelp.get_classifier_id_for_prediction():
            classifier_kwargs = yhelp.make_classifier_args(clsfr_id)
            predict_links(**classifier_kwargs)

    if yhelp.do_upload():
        upload_kwargs = yhelp.make_upload_args()
        upload_dir_to_s3(**upload_kwargs)

    return None


@cli.command()
@click.option("--input_path", nargs=1, help="The path to the yaml to update.")
@click.option(
    "--keys",
    callback=lambda _, __, x: x.split(",") if x else [],
    help="One or more keys to update the values for, comma-delimited. "
    "Nested keys (i.e., keys under other keys) must be delimited "
    "with colons, e.g. key1:key2:key3.",
)
@click.option(
    "--values",
    callback=lambda _, __, x: x.split(",") if x else [],
    help="One or more values, in the same order as keys, comma-delimited.",
)
def updateyaml(input_path, keys, values):
    """Update a YAML file with specified key/value pairs
    \f
    Updates one or more values for a one or more keys,
    with a provided path to a YAML file.
    Will not replace keys found multiple times in the YAML.
    Ignores keys in lists, even if they're dicts in lists.
    """
    do_update_yaml(input_path, keys, values)



import functools
import logging

import os
import pickle
import string
from typing import Optional, Union
from urllib.request import Request, urlopen
import tarfile
import sys
import pkg_resources # type: ignore

import yaml  # type: ignore
from grape import Graph # type: ignore
from neat_ml.link_prediction.mlp_model import MLPModel  # type: ignore
from neat_ml.link_prediction.model import Model
import validators  # type: ignore
from pathlib import Path

import neat_ml_schema # type: ignore
from linkml_validator.validator import Validator # type: ignore

from neat_ml.run_classifier.run_classifier import get_custom_model_path

VALID_CHARS = "-_.() %s%s" % (string.ascii_letters, string.digits)


def validate_config(config: dict, neat_schema_file: str = 'neat_ml_schema.yaml') -> bool:
    """
    Validate the provided config against the neat_schema.
    :param config: dict of the parsed config file
    :param neat_schema_file: name of the neat schema file in neat-schema package
    :return: bool, false if validation failed
    """
    validated = True

    # Get schema path first
    schema_path = pkg_resources.resource_filename('neat_ml_schema',
                                                  os.path.join('src/schema/',
                                                               neat_schema_file))

    if not os.path.exists(schema_path):
        raise RuntimeError

    validator = Validator(schema=schema_path)
    for class_type in config:
        try:
            validator.validate(obj=config, target_class=class_type)
        except (KeyError, ValueError):
            validated = False
            print(f"Config failed validation for {class_type}")

    return validated


def parse_yaml(file: str) -> dict:
    with open(file, "r") as stream:
        return yaml.load(stream, Loader=yaml.FullLoader)

def is_url(string_to_check: Union[str, Path]) -> bool:
    """Helper function to decide if a string is a
    URL (used for example for deciding
    whether we need to download a file for a given node_path or edge_path).
    Raise exception if file path is invalid.
    :param string_to_check: string to check
    :return: bool, True if string is URL
    """

    return bool(validators.url(string_to_check))


def is_valid_path(string_to_check: Union[str, Path]) -> bool:
    """Helper function to decide if a string is a
    invalid filepath.
    Raise exception if file path is invalid.
    :param string_to_check: string to check
    :return: bool, True if string is valid filepath
    """

    if isinstance(string_to_check, Path):
        if not string_to_check.is_file():
            raise FileNotFoundError(
                f"{string_to_check} is not a valid file path or url."
            )
    elif not os.path.exists(string_to_check):
        raise FileNotFoundError(
            f"{string_to_check} is not a valid file path or url."
        )
    else:
        return True

    return False

def download_file(url: str, outfile: str) -> list:
    """
    Downloads file at input url to outfile path.
    URL must point to a TSV or a tar.gz compressed file.
    (This is checked during pre_run_checks though.)
    If it's tar.gz, decompress.
    Return the names of all files as a list.
    """

    outlist = []

    req = Request(url, headers={'User-Agent': 'Mozilla/5.0'})
    with urlopen(req) as response, open(outfile, 'wb') as fh:  # type: ignore
        data = response.read()  # a `bytes` object
        fh.write(data)
    if outfile.lower().endswith(".tar.gz"): # Need to decompress
        decomp_outfile = tarfile.open(outfile)
        outdir = os.path.dirname(outfile)
        for filename in decomp_outfile.getnames():
            outlist.append(os.path.join(outdir,filename))
        filecount = len(outlist)
        if filecount > 2:
            logging.warning(f"{outfile} contains {filecount} files.")
        decomp_outfile.extractall(outdir)
        decomp_outfile.close()
    else:
        outlist.append(outfile)

    return outlist


def catch_keyerror(f):
    @functools.wraps(f)
    def func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except KeyError as e:
            print(f"Can't find key in YAML: {e} (possibly harmless)")
            return None

    return func


class YamlHelper:
    """
    Class to parse yaml and extract args for methods
    """

    def __init__(self, config: str):
        self.default_outdir = "output_data"
        self.default_indir = ""
        self.yaml: dict = parse_yaml(config)

        if not validate_config(self.yaml):
            sys.exit("Please check config file! Exiting...")

    def indir(self):
        """Get input directory.
            This is not currently specified in the
            config and will be the current
            working directory unless changed
            directly through a call to the
            YamlHelper object.

        Returns:
            The input directory

        """
        indir = self.default_indir

        return indir

    def outdir(self):
        """Get output directory from config.

        Returns:
            The output directory

        """
        outdir = (
            self.yaml["Target"]["target_path"]
            if "Target" in self.yaml
            else self.default_outdir
        )
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        return outdir

    def retrieve_from_sources(self) -> None:
        """
        Checks for existence of a
        source_data key. If this exists,
        download and decompress as needed.
        The node_path and edge_path values
        in graph_data
        will still need to refer to the
        node/edge filenames.
        """

        if "GraphDataConfiguration" in self.yaml:
            if "source_data" in self.yaml["GraphDataConfiguration"]:
                for entry in self.yaml["GraphDataConfiguration"]["source_data"]["files"]:
                    filepath = entry["path"]
                    if "desc" in entry:
                        desc = entry["desc"]
                        print(f"Retrieving {filepath}: {desc}")
                    else:
                        print(f"Retrieving {filepath}")
                    if is_url(filepath):
                        url_as_filename = \
                            ''.join(c if c in VALID_CHARS else "_" for c in filepath)
                        outfile = os.path.join(self.indir(), url_as_filename)
                        download_file(filepath, outfile)
                    # If this was a URL, it already got decompressed.
                    # but if it's local and still compressed, decompress now
                    # (this can happen if we already downloaded it but didn't decomp)
                    if filepath.endswith(".tar.gz"):
                        outlist = []
                        if is_url(filepath):
                            decomp_outfile = tarfile.open(outfile)
                        else:
                            decomp_outfile = tarfile.open(filepath)
                        for filename in decomp_outfile.getnames():
                            outlist.append(os.path.join(self.indir(),filename))
                        decomp_outfile.extractall(self.indir())
                        decomp_outfile.close()

    def add_indir_to_graph_data(
        self,
        graph_data: dict,
        keys_to_add_indir: list = ["node_path", "edge_path"],
    ) -> dict:
        """
        Updates the graph file paths
        with their input directory.
        :param graph_data - parsed yaml
        :param keys_to_add_indir: what keys to add indir to
        :return:
        """

        for k in keys_to_add_indir:
            if k in graph_data:
                graph_data[k] = os.path.join(self.indir(), graph_data[k])
            else:
                logging.warning(
                    f"Can't find key {k} in graph_data - skipping (possibly harmless)"
                )
        return graph_data

    #
    # graph stuff
    #
    def load_graph(self) -> Graph:
        """
        Loads graph nodes and edges into Ensmallen.
        Creates a node type list, as Ensmallen
        requires this to parse node types.
        :return: ensmallen Graph
        """

        #Load sources if necessary
        self.retrieve_from_sources()

        graph_args_with_indir = self.main_graph_args()

        for pathtype in ["node_path", "edge_path"]:
            filepath = graph_args_with_indir[pathtype]
            if is_url(filepath):
                url_as_filename = "".join(
                    c if c in VALID_CHARS else "_" for c in filepath
                )
                outfile = os.path.join(self.outdir(), url_as_filename)
                download_file(filepath, outfile)
                graph_args_with_indir[pathtype] = os.path.join(outfile)
            elif not is_valid_path(filepath):
                raise FileNotFoundError(f"Please check path: {filepath}")

        # Now load the Ensmallen graph
        loaded_graph = Graph.from_csv(**graph_args_with_indir)

        return loaded_graph

    def main_graph_args(self) -> dict:
        return self.add_indir_to_graph_data(self.yaml["GraphDataConfiguration"]["graph"])

    @catch_keyerror
    def val_graph_args(self) -> dict:
        return self.add_indir_to_graph_data(
            self.yaml["GraphDataConfiguration"]["evaluation_data"]["valid_data"]
        )

    @catch_keyerror
    def train_graph_args(self) -> dict:
        return self.add_indir_to_graph_data(
            self.yaml["GraphDataConfiguration"]["evaluation_data"]["train_data"]
        )

    #
    # embedding stuff
    #

    def do_embeddings(self) -> bool:
        return "EmbeddingsConfig" in self.yaml

    def embedding_outfile(self) -> str:
        filepath = self.yaml['EmbeddingsConfig']['filename']
        if is_url(filepath):
            url_as_filename = \
                ''.join(c if c in VALID_CHARS else "_" for c in filepath)
            outfile = os.path.join(self.outdir(), url_as_filename)
            download_file(filepath, outfile)
            return outfile
        else:
            return os.path.join(self.outdir(),
                            filepath)

    @catch_keyerror
    def embedding_history_outfile(self):
        return os.path.join(
            self.outdir(),
            self.yaml["EmbeddingsConfig"]["history_filename"],
        )

    # def make_embeddings_metrics_class_list(self) -> list:
    #     metrics_class_list = []

    #     metrics = (
    #         self.yaml["EmbeddingsConfig"]["metrics"]
    #         if "metrics" in self.yaml["embeddings"]
    #         else None
    #     )
    #     if metrics:
    #         for m in metrics:
    #             if m["type"].startswith("tensorflow.keras"):
    #                 m_class = Model.dynamically_import_class(m["type"])
    #                 m_parameters = m["parameters"]
    #                 m_instance = m_class(**m_parameters)  # type: ignore
    #                 metrics_class_list.append(m_instance)
    #             else:
    #                 metrics_class_list.append([m["type"]])
    #     return metrics_class_list

    def make_node_embeddings_args(self) -> dict:
        node_embedding_args = {
            "embedding_outfile": self.embedding_outfile(),
            "embedding_history_outfile": self.embedding_history_outfile(),
            "main_graph_args": self.main_graph_args(),
            "node_embedding_params": self.yaml["EmbeddingsConfig"][
                "node_embeddings_params"
            ],
            "bert_columns": self.yaml["EmbeddingsConfig"]["bert_params"][
                "node_columns"
            ]
            if "bert_params" in self.yaml["EmbeddingsConfig"]
            else None,
        }
        return node_embedding_args

    #
    # tSNE stuff
    #

    def do_tsne(self) -> bool:
        return "tsne_filename" in self.yaml["EmbeddingsConfig"]

    def make_tsne_args(self, graph: Graph) -> dict:
        make_tsne_args = {
            "graph": graph,
            "tsne_outfile": self.tsne_outfile(),
            "embedding_file": self.embedding_outfile(),
        }
        return make_tsne_args

    def tsne_outfile(self) -> str:
        return os.path.join(
            self.outdir(), self.yaml["EmbeddingsConfig"]["tsne_filename"]
        )

    #
    # classifier stuff
    #

    def do_classifier(self) -> bool:
        return "ClassifierContainer" in self.yaml

    def classifier_type(self) -> str:
        return self.yaml["ClassifierContainer"]["classifiers"]["classifier_name"]

    @catch_keyerror
    def classifiers(self) -> list:
        """From the YAML, extract a list of classifiers to be trained

        :return: list of classifiers to be trained
        """
        return self.yaml["ClassifierContainer"]["classifiers"]

    def get_all_classifier_ids(self):
        return [c["classifier_id"] for c in self.yaml["ClassifierContainer"]["classifiers"]]

    def get_edge_embedding_method(self, classifier: dict) -> str:
        return classifier["edge_method"]

    def classifier_history_file_name(self, classifier: dict) -> Optional[str]:
        return (
            classifier["history_filename"]
            if "history_filename" in classifier
            else None
        )

    def classifier_outfile(self, classifier: dict) -> str:
        return os.path.join(
            self.outdir(), classifier["outfile"]
        )

    #
    # upload stuff
    #

    def do_upload(self) -> bool:
        return "Upload" in self.yaml

    def make_upload_args(self) -> dict:
        make_upload_args = {
            "local_directory": self.outdir(),
            "s3_bucket": self.yaml["Upload"]["s3_bucket"],
            "s3_bucket_dir": self.yaml["Upload"]["s3_bucket_dir"],
            "extra_args": self.yaml["Upload"]["extra_args"]
            if "extra_args" in self.yaml["Upload"]
            else None,
        }
        return make_upload_args

    #
    # applying trained model to fresh data for predictions
    #
    def do_apply_classifier(self):
        return "ApplyTrainedModelsContainer" in self.yaml

    def get_classifier_id_for_prediction(self):
        classifier_applications = self.yaml["ApplyTrainedModelsContainer"]["models"]
        list_of_ids = [
            cl["model_id"] for cl in classifier_applications
        ]
        return list_of_ids

    def get_classifier_from_id(self, classifier_id: str):

        return [
            x
            for x in self.classifiers()
            if x["classifier_id"] == classifier_id
        ][0]

    def make_classifier_args(self, cl_id: str):
        classifier_args = [
            arg
            for arg in self.yaml["ApplyTrainedModelsContainer"]["models"]
            if cl_id == arg["model_id"]
        ][0]
        model = self.get_classifier_from_id(cl_id)

        classifier_args_dict = {}
        classifier_args_dict["graph"] = Graph.from_csv(
            **self.main_graph_args()
        )

        if self.get_classifier_from_id(cl_id)["classifier_name"] == "neural network":
            classifier_args_dict["model"] = pickle.load(
                open(
                    os.path.join(
                        self.outdir(),
                        get_custom_model_path(model["outfile"]),
                    ),
                    "rb",
                ),
            )

        else:
            classifier_args_dict["model"] = pickle.load(
                open(
                    os.path.join(self.outdir(), model["outfile"]),
                    "rb",
                ),
            )

        # YAML may specify node_types as dict with 'source' and 'destination'
        # or as a list of lists
        if 'source' in classifier_args["node_types"] \
            or 'destination' in classifier_args["node_types"]:
            classifier_args_dict["node_types"] = [classifier_args["node_types"]['source'],
                                                    classifier_args["node_types"]['destination']]
        else:
            classifier_args_dict["node_types"] = classifier_args["node_types"]

        classifier_args_dict["cutoff"] = classifier_args["cutoff"]
        classifier_args_dict["output_file"] = os.path.join(
            self.outdir(), classifier_args["outfile"]
        )

        classifier_args_dict["embeddings_file"] = self.embedding_outfile()
        classifier_args_dict["edge_method"] = model["edge_method"]

        return classifier_args_dict



import collections
import warnings

import boto3  # type: ignore
from botocore.exceptions import ClientError  # type: ignore
from neat_ml.yaml_helper.yaml_helper import YamlHelper

def pre_run_checks(
    yhelp: YamlHelper,
    check_s3_credentials: bool = True,
    check_s3_bucket: bool = True,
    check_s3_bucket_dir: bool = True,
    check_classifiers: bool = True,
) -> bool:
    """Some checks before run, to prevent frustrating failure at the end of long runs

    Args:
        yhelp: YamlHelper object
        check_s3_credentials: should we check S3 credentials (true). Note that if no
            upload dir exists, this will pass
        check_s3_bucket: check that s3 bucket exists on s3
        check_s3_bucket_dir: check that s3 bucket directory doesn't already exist
        check_classifiers: verify that classifier ids don't conflict

    Returns:
        Boolean pass or fail
    """
    return_val = True

    if check_s3_credentials:
        try:
            client = boto3.client("s3")
            client.list_buckets()  # to check credentials
        except ClientError as ce:
            warnings.warn(f"Client error when trying S3 credentials: {ce}")
            if yhelp.do_upload():
                return_val = False
            else:
                warnings.warn("YAML contains no upload block - continuing")

    if (
        check_s3_bucket and yhelp.do_upload()
    ):  # make sure we are going to upload
        upload_args = yhelp.make_upload_args()
        try:
            client = boto3.client("s3")
            buckets = []
            bucket_info = client.list_buckets()
            if "Buckets" in bucket_info:
                buckets = [
                    this_dict["Name"] for this_dict in bucket_info["Buckets"]
                ]
            else:
                warnings.warn(
                    "Can't find 'Buckets' key in output of client.list_buckets()"
                )
            if "s3_bucket" not in upload_args:
                warnings.warn("No 's3_bucket' in upload block")
                return_val = False
            elif upload_args["s3_bucket"] not in buckets:
                warnings.warn(
                    f"Can't find the target s3 bucket "
                    f"{upload_args['s3_bucket']} in s3"
                )
                return_val = False
        except ClientError as ce:
            warnings.warn(f"Client error when trying S3 credentials: {ce}")
            return_val = False

    if (
        check_s3_bucket_dir and yhelp.do_upload()
    ):  # make sure we are going to upload
        upload_args = yhelp.make_upload_args()
        
        if not pre_bucket_check(upload_args):
            return_val = False

    if check_classifiers and yhelp.classifiers():
        # PSEUDOCODE
        # if classifiers:
        # Then make sure id-s aren't clashing
        # &
        # classifier_model_id is calling a valid id.

        all_classifier_ids = yhelp.get_all_classifier_ids()
        if len(all_classifier_ids) != len(set(all_classifier_ids)):
            dup_ids = [
                item # type: ignore
                for item, count in collections.Counter(  # type: ignore
                    all_classifier_ids
                ).items()
                if count > 1
            ]

            raise ValueError(
                f"Same 'classifier_id' represents multiple classes in the yaml provided: {dup_ids}"
            )

        if yhelp.do_apply_classifier():
            check = all(
                item in all_classifier_ids
                for item in yhelp.get_classifier_id_for_prediction()
            )
            if not check:
                return_val = False
                raise ValueError(
                    f"The 'classifier_id' used for prediction does "
                    "not map to any classifier in the yaml provided:"
                    "{yhelp.get_classifier_id_for_prediction()}"
                )

    return return_val

def pre_bucket_check(upload_args: dict) -> bool:
    """
    Given upload args, checks if the
    target bucket and directory are
    accessible and empty, respectively.
    """

    success = True

    try:
        client = boto3.client("s3")
        if "s3_bucket_dir" not in upload_args:
            warnings.warn("No 's3_bucket_dir' in upload block")
            success = False
        else:
            result = client.list_objects(
                Bucket=upload_args["s3_bucket"],
                Prefix=upload_args["s3_bucket_dir"],
            )
            if "Contents" in result:
                warnings.warn(
                    f"There are already objects in remote s3 directory: "
                    f"{upload_args['s3_bucket_dir']}"
                )
                success = False
    except ClientError as ce:
        warnings.warn(f"Client error when trying S3 credentials: {ce}")
        success = False

    return success

if __name__ == "__main__":
    pre_run_checks(yhelp=YamlHelper())  # type: ignore



import os
import pickle

import numpy as np  # type: ignore
import copy
import pandas as pd  # type: ignore
from typing import Optional, Tuple, Union
from grape.embedding_transformers import EdgePredictionTransformer, GraphTransformer  # type: ignore
from grape import Graph  # type: ignore
import importlib


class Model:
    def __init__(self, outdir=None):
        if outdir:
            os.makedirs(outdir, exist_ok=True)
        self.outdir = outdir

    def fit(self, X, y):
        pass

    def compile(self):
        pass

    def load(self, path: str) -> object:
        pass

    def save(self):
        with open(
            os.path.join(self.outdir, self.config["outfile"]), "wb"
        ) as f:
            pickle.dump(self, f)

    def predict(self, predict_data) -> np.ndarray:
        return self.model.predict(predict_data) # type: ignore

    def predict_proba(self, X) -> np.ndarray:
        pass

    @classmethod
    def make_train_valid_data(
        self,
        embedding_file: str,
        training_graph_args: dict,
        validation_args: Optional[dict] = None,
        training_args: Optional[dict] = None,
        edge_method: str = "Average",
    ) -> Tuple[Tuple, Tuple]:
        """Prepare training and validation data for training link prediction classifers

        Args:
            embedding_file: path to embedding file for nodes in graph
            training_graph_args: Ensmallen arguments to load training graph
            validation_args: Ensmallen arguments to load positive and negative validation graphs
            training_args: Ensmallen arguments to load negative training graph
            edge_method: edge embedding method to use (average, L1, L2, etc)
        Returns:
            A tuple of tuples

        """

        embedding = pd.read_csv(embedding_file, index_col=0, header=None)

        # load graphs
        graphs = {"pos_training": Graph.from_csv(**training_graph_args)}
        is_directed = graphs["pos_training"].is_directed()
        graph_filepaths = {
            "pos_validation": "pos_edge_filepath",
            "neg_training": "neg_edge_filepath",
            "neg_validation": "neg_edge_filepath"
        }
        for name, graph_args in [
            ("pos_validation", validation_args),
            ("neg_training", training_args),
            ("neg_validation", validation_args),
        ]:
            if not graph_args:
                if name in ["neg_training", "neg_validation"]:
                    neg_edge_number = graphs["pos_training"].get_edges_number()
                    if not is_directed:
                        neg_edge_number = neg_edge_number * 2

                    graphs[name] = (graphs["pos_training"]).sample_negative_graph(
                        number_of_negative_samples=neg_edge_number
                    )
                else:
                    these_params = copy.deepcopy(training_graph_args)
                    if "directed" not in these_params.keys():
                        these_params["directed"] = training_graph_args[
                            "directed"
                        ]
                    graphs[name] = Graph.from_csv(**these_params)
            else:
                these_params = copy.deepcopy(training_graph_args)
                these_params["edge_path"] = graph_args[graph_filepaths[name]]
                graphs[name] = Graph.from_csv(**these_params)
                                                    
        # create transformer object to convert graphs into edge embeddings
        lpt = EdgePredictionTransformer(method=edge_method)

        lpt.fit(
            embedding
        )  # pass node embeddings to be used to create edge embeddings

        train_edges, train_labels = lpt.transform(
            positive_graph=graphs["pos_training"],
            negative_graph=graphs["neg_training"],
        )
        valid_edges, valid_labels = lpt.transform(
            positive_graph=graphs["pos_validation"],
            negative_graph=graphs["neg_validation"],
        )
        return (train_edges, train_labels), (valid_edges, valid_labels)

    @classmethod
    def make_edge_embedding_for_predict(
        self,
        embedding_file: str,
        edge_method: str,
        source_destination_list
    ) -> np.ndarray:
        """Prepare training and validation data for training link prediction classifers

        Args:
            embedding_file: path to embedding file for nodes in graph
            trained_graph_args: EnsmallenGraph arguments to load training graph
            edge_method: edge embedding method to use (average, L1, L2, etc)
        Returns:
            A NumPy Array embeddings that represent prediction edges.

        """

        embedding = pd.read_csv(embedding_file, index_col=0, header=None)

        # Create graphtransformer object for edge embeddings
        gt = GraphTransformer(method=edge_method)

        gt.fit(embedding)

        edge_embedding_for_predict = gt.transform(
            graph=source_destination_list
        )

        return edge_embedding_for_predict

    @classmethod
    def dynamically_import_class(self, reference) -> object:
        """Dynamically import a class based on its reference.

        Args:
            reference: The reference or path for the class to be imported.

        Returns:
            The imported class

        """
        klass = self.my_import(reference)
        return klass

    @classmethod
    def dynamically_import_function(self, reference) -> object:
        """Dynamically import a function based on its reference.

        Args:
            reference: The reference or path for the function to be imported.

        Returns:
            The imported function

        """
        module_name = ".".join(reference.split(".")[0:-1])
        function_name = reference.split(".")[-1]
        f = getattr(importlib.import_module(module_name), function_name)
        return f

    @classmethod
    def my_import(self, name):
        components = name.split(".")
        mod = __import__(components[0])
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod



import pandas as pd  # type: ignore
from grape import Graph  # type: ignore
from grape import GraphVisualizer  # type: ignore


def make_tsne(
        graph: Graph,
        tsne_outfile: str,
        embedding_file: str,
) -> None:
    """
    Generate a simple node tSNE plot.
    :param graph: Ensmallen graph object
    :param tsne_outfile: where to output tSNE file
    :param embedding_file: file containing embeddings (in numpy format) to use to
        generate the tSNE plot
    :return:
    """

    node_embeddings = pd.read_csv(embedding_file, index_col=0, header=None)

    visualizer = GraphVisualizer(graph)
    figure = visualizer.plot_nodes(visualizer.fit_nodes(node_embeddings))
    print(figure[1])
    figure[0].savefig(tsne_outfile)

def make_all_plots(
        graph: Graph,
        tsne_outfile: str,
        embedding_file: str,
        ) -> None:
    """
    Generate the full set of Ensmallen plots.
    :param graph: Ensmallen graph object
    :param tsne_outfile: where to output plot image file
    :param embedding_file: file containing embeddings (in numpy format) to use to
        generate the tSNE plot
    :return:
    """

    node_embeddings = pd.read_csv(embedding_file, index_col=0, header=None)

    visualizer = GraphVisualizer(graph)
    figure = visualizer.fit_and_plot_all(node_embeddings)
    print(figure[1])
    figure[0].savefig(tsne_outfile)


import os
import pickle
import tensorflow as tf  # type: ignore
from .model import Model


class MLPModel(Model):
    def __init__(self, config, outdir: str = None) -> None:
        """make an MLP model

        Args:
            config: the classifier config

        Returns:
            The model

        """
        super().__init__(outdir=outdir)
        self.config = config
        model_type = config["classifier_type"]
        model_class = self.dynamically_import_class(model_type)
        model_layers = []
        for layer in config["parameters"]["tf_keras_params"]["layers_config"]["layers"]:
            layer_type = layer["type"]
            layer_class = self.dynamically_import_class(layer_type)
            parameters = layer["parameters"]
            layer_instance = layer_class(**parameters)  # type: ignore
            model_layers.append(layer_instance)
        model_instance = model_class()  # type: ignore
        for l in model_layers:
            model_instance.add(l)
        self.model = model_instance

    def compile(self):
        model_compile_parameters = self.config["parameters"]["tf_keras_params"]
        metrics = (
            model_compile_parameters["metrics_config"]["metrics"]
            if "metrics_config" in model_compile_parameters
            else None
        )
        metrics_class_list = []
        for m in metrics:
            if m["type"].startswith("tensorflow.keras"):
                m_class = self.dynamically_import_class(m["type"])
                m_parameters = m["parameters"]
                m_instance = m_class(**m_parameters)
                metrics_class_list.append(m_instance)
            else:
                metrics_class_list.append([m["type"]])
        self.model.compile(
            loss=model_compile_parameters["loss"],
            optimizer=model_compile_parameters["optimizer"],
            metrics=metrics_class_list,
        )

    def fit(self, train_data, train_labels):
        """Takes a model, generated from make_model(), and calls .fit()

        Args:
            train_data: training data for fitting
            validation_data: validation data for fitting

        Returns:
            The model object

        """
        try:
            classifier_params = self.config["parameters"]["tf_keras_params"][
                "fit_config"
            ]
        except KeyError:
            classifier_params = {}

        callback_list = []
        if "callbacks_list" in classifier_params:
            for callback in classifier_params["callbacks_list"]["callbacks"]:
                c_class = self.dynamically_import_class(callback["type"])
                c_params = (
                    callback["parameters"] if "parameters" in callback else {}
                )
                c_instance = c_class(**c_params)
                callback_list.append(c_instance)
            del classifier_params["callbacks"]

        history = self.model.fit(
            train_data,
            train_labels,
            **classifier_params,
            callbacks=callback_list
        )
        return history

    def save(self) -> None:
        self.model.save(
            os.path.join(self.outdir, self.config["outfile"])
        )

        fn, ext = os.path.splitext(self.config["outfile"])
        model_outfile = fn + "_custom" + ext

        with open(os.path.join(self.outdir, model_outfile), "wb") as f:
            pickle.dump(self, f)

    def load(self, path: str) -> tuple(): # type: ignore
        fn, ext = os.path.splitext(path)
        custom_model_filename = fn + "_custom" + ext
        generic_model_object = tf.keras.models.load_model(path)

        with open(custom_model_filename, "rb") as mf2:
            custom_model_object = pickle.load(mf2)

        return generic_model_object, custom_model_object



import os
import re

from codecs import open as copen  # to use a consistent encoding
from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

# get the long description from the relevant file
with copen(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def read(*parts):
    with copen(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


__version__ = "0.1.0"

test_deps = [
    "pytest",
    "pytest-cov",
    "coveralls",
    "validate_version_code",
    "codacy-coverage",
    "parameterized",
    "mypy",
]

extras = {
    "test": test_deps,
}

setup(
    name="neat_ml",
    version=__version__,
    description="Neural-network Embedding All the Things",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/Knowledge-Graph-Hub/neat_ml",
    author="Deepak Unni, Justin Reese, J. Harry Caufield, Harshad Hegde",
    author_email="deepak.unni3@lbl.gov, justaddcoffee+github@gmail.com, jhc@lbl.gov, hhegde@lbl.gov",
    python_requires=">=3.8",
    license="BSD-3",
    include_package_data=True,
    classifiers=[],
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    entry_points={"console_scripts": ["neat=neat_ml.cli:cli"]},
    tests_require=test_deps,
    install_requires=[
        # tensorflow and torch may also be required,
        # but should be installed separately
        # to avoid version conflicts
        "tqdm",
        "click",
        "pyyaml",
        "sklearn",
        "grape",
        "opencv-python",  # for embiggen's 4d tSNEs
        "numpy",
        "pandas",
        "transformers",
        "boto3",
        "botocore",
        "validators",
        "linkml",
        "neat-ml-schema",
        "linkml-validator"
    ],
    extras_require=extras,
)



import os
from pathlib import Path
from typing import List, Union
from warnings import warn

from grape import Graph  # type: ignore
import pandas as pd  # type: ignore
from itertools import combinations  # [READ DOCS]

from neat_ml.link_prediction.sklearn_model import SklearnModel

OUTPUT_COL_NAMES = ["source_node", "destination_node"]


def gen_src_dst_pair(
    graph: Graph,
    ignore_existing_edges: bool = True,
):

    # Get all node ids
    node_ids = graph.get_node_ids().tolist()[:100]
    # Yield only the (src, dst) combo
    # that does NOT exist in the graph.
    for combo in list(combinations(node_ids, 2)):
        # * 4 cases
        # 1. Graph DIRECTED & IGNORE Existing edges
        # 2. Graph DIRECTED & INCLUDE Existing edges
        # 3. Graph UNDIRECTED & IGNORE Existing edges
        # 4. Graph UNDIRECTED & INCLUDE Existing edges

        # Graph DIRECTED
        if graph.is_directed():
            # IGNORE Existing edges: Neither combo exist.
            if ignore_existing_edges:
                if not graph.has_edge_from_node_ids(
                    *combo
                ) and not graph.has_edge_from_node_ids(
                    *tuple(reversed(combo))
                ):
                    yield combo
            # INCLUDE Existing edges: yield every combo
            else:
                yield combo

        # Graph UNDIRECTED
        else:
            # IGNORE Existing edges:
            if ignore_existing_edges:
                if not graph.has_edge_from_node_ids(
                    *combo
                ) or not graph.has_edge_from_node_ids(*tuple(reversed(combo))):
                    yield combo
            # INCLUDE Existing edges: yield every combo
            else:
                yield combo


def predict_links(
    graph: Graph,
    model: object,
    node_types: List[List],
    cutoff: float,
    output_file: Union[str, Path],
    embeddings_file: Union[str, Path],
    edge_method: str,  # [Average etc.]
    ignore_existing_edges: bool = True,
    verbose: bool = True,
) -> None:
    """Performs link prediction over provided graph nodes.

    Args:
        graph (Graph): Ensmallen graph.
        model (Any): Trained model.
        node_types (list): List of lists of target 'source' and 'destination' nodes.
                            Only these types will be in output.
        cutoff (float): Cutoff point for filtering.
        output_file (str or Path): Results destination.
        embeddings_file (str or Path): Path to embeddings.
        edge_method (str): Method to use for calculating edge embeddings.
        ignore_existing_edges (bool): default True; do not output predictions for edges already in graph.
    """
    embeddings = pd.read_csv(embeddings_file, sep=",", header=None)

    embedding_node_names = list(embeddings[0])
    src_dst_list = []
    no_embed_list = []

    for src, dst in gen_src_dst_pair(graph, ignore_existing_edges):

        src_name = graph.get_node_name_from_node_id(src)
        dst_name = graph.get_node_name_from_node_id(dst)

        # Check if this pair passes the node filter(s)
        # Note that nodes may have >1 type
        # so these are lists
        if node_types:
            src_types = graph.get_node_type_names_from_node_id(src)
            dst_types = graph.get_node_type_names_from_node_id(dst)

            # Test if any intersection between node_types and src/dst types
            if len(list(set(src_types) & set(node_types[0]))) == 0 or \
                len(list(set(dst_types) & set(node_types[1]))) == 0 :
                continue

        # see if src and dst are actually in embedding.tsv
        for name in [src_name, dst_name]:
            if name not in embedding_node_names:
                if verbose:
                    warn(f"Can't find {name} in embeddings - skipping")
                no_embed_list.append((src_name, dst_name))
            else:
                src_dst_list.append((src_name, dst_name))

    if len(src_dst_list) == 0:
        warn("Filter has excluded all edges or no edges found - cannot apply classifier.")

    edge_embedding_for_predict = model.make_edge_embedding_for_predict(  # type: ignore
        embedding_file=embeddings_file,  # this should be the new embeddings
        edge_method=edge_method,
        source_destination_list=src_dst_list,
    )

    embed_df = pd.DataFrame(src_dst_list, columns=OUTPUT_COL_NAMES)
    # NOTE: A trained Sklearn model treats '0' and '1' labels as classes
    #  as opposed to a Tensorflow(MLP) model where 0 and 1 are booleans
    # to a class (binary).

    if type(model) == SklearnModel:
        pred_probas = [
            y for x, y in model.predict_proba(edge_embedding_for_predict)
        ]
        pred_proba_df = pd.DataFrame(pred_probas, columns=["score"])
        full_embed_df = pd.concat([embed_df, pred_proba_df], axis=1)
    else:
        preds = model.predict(edge_embedding_for_predict)  # type: ignore
        embed_df["score"] = preds
        full_embed_df = embed_df

    if no_embed_list:
        no_embed_df = pd.DataFrame(no_embed_list, columns=OUTPUT_COL_NAMES)
        output_df = pd.concat([full_embed_df, no_embed_df], axis=1)
    else:
        output_df = full_embed_df

    if cutoff:
        filtered_output = output_df[output_df["score"] > cutoff]
        output_df = filtered_output

    output_df.sort_values(by="score", inplace=True, ascending=False)
    output_df.to_csv(output_file, sep="\t", index=None)


# This may be moved if needed
def get_custom_model_path(model_file_path: str) -> str:
    """
    Given the path to a sklearn or TF model,
    returns the name of the corresponding custom
    model. This allows a NEAT Model object to be
    created so we may access its methods.
    :param model_file_path: str, path to generic model
    :return: str, path to custom model
    """

    custom_model_path = (
        os.path.splitext(model_file_path)[0]
        + "_custom"
        + os.path.splitext(model_file_path)[1]
    )
    return custom_model_path



import yaml

def do_update_yaml(input_path: str, keys: list, values: list) -> None:
    with open(input_path, 'r') as yaml_file:
        contents = yaml.load(yaml_file, Loader=yaml.FullLoader)

    newkeyvalues = tuple(zip(keys, values))
    for key, newvalue in newkeyvalues:
        print(f"Will set \"{key}\" to \"{newvalue}\".")

        try:
            # Parse the input key
            if len(key.split(":")) == 1:
                contents = update_keyvalue(contents, [key], newvalue)
            else:
                keylist = key.split(":")
                contents = update_keyvalue(contents, keylist, newvalue)
        except KeyError:
            print(f"Could not find \"{key}\"! Skipping.")
            continue

        print("Done.")

    with open(input_path, 'w') as yaml_file:
        yaml_file.write(yaml.dump(contents, default_flow_style=False, sort_keys=False))


def update_keyvalue(input_dict, keys, newvalue):
    """
    Function to update a provided key with a value.
    :param input_dict: YAML dict representation to be updated
    :param keys: list containing one or more keys. 
                    Multiple keys are nested, e.g. input_dict[key1][key2[key3].
    :param newvalue: value to set key to.
    :return: dict
    """
    new_dict = input_dict.copy()

    if len(keys) == 1 and keys[0] in new_dict:
        new_dict[keys[0]] = newvalue
        return new_dict
    else:
        i = 0
        current_dict = new_dict[keys[i]]
        while True:
            if i == len(keys)-2:
                current_dict[keys[i+1]] = newvalue
                return new_dict
            else:
                i = i+1
                current_dict = current_dict[keys[i]]
                
                




import transformers # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from tqdm.auto import tqdm  # type: ignore
from grape import Graph  # type: ignore
from grape.embedders import embed_graph # type: ignore

def get_node_data(file: str, sep="\t") -> pd.DataFrame:
    """Read node TSV file and return pandas dataframe

    :param file: node TSV file
    :param sep: separator
    :return:
    """
    return pd.read_csv(file, sep=sep)


def make_node_embeddings(
                         embedding_outfile: str,
                         embedding_history_outfile: str,
                         main_graph_args: dict,
                         node_embedding_params: dict,
                         bert_columns: dict,
                         bert_pretrained_model: str = "allenai/scibert_scivocab_uncased"
                         ) -> None:
    """Make embeddings and output embeddings and training history

    Args:
        embedding_outfile: outfile to write out embeddings
        main_graph_args: arguments passed to Ensmallen for graph loading
        node_embedding_params: args passed to Embiggen
        bert_columns: columns containing text info to use to make embeddings from Bert
                pretrained embeddings
    Returns:
        None.

    """
    # load main graph
    graph: Graph = Graph.from_csv(**main_graph_args)

    node_embedding = embed_graph(
        graph,
        node_embedding_params["method_name"],
        node_embedding_params
    ).get_node_embedding_from_index(0)

    # embed columns with BERT first (if we're gonna)
    bert_embeddings = pd.DataFrame()
    if bert_columns:
        bert_model = transformers.BertModel.from_pretrained(bert_pretrained_model,
                                               output_hidden_states=True)
        bert_tokenizer = transformers.BertTokenizer.from_pretrained(bert_pretrained_model)
        bert_model.eval()
        all_bert_embeddings = bert_model.embeddings.word_embeddings.weight.data.numpy()

        node_data = get_node_data(main_graph_args['node_path'])

        node_text = [
            " ".join([str(row[col]) for col in bert_columns])
            for index, row in tqdm(node_data.iterrows(), "extracting text from nodes")
        ]
        node_text_tokenized = [bert_tokenizer.encode(
            this_text,  # Sentence to encode
            # add_special_tokens=True,  # Add '[CLS]' and '[SEP]'
            return_tensors='np'
        ) for this_text in tqdm(node_text, "tokenzing text")]
        node_text_tensors = [
            np.mean(all_bert_embeddings[ids.flatten()], axis=0)
            for ids in tqdm(node_text_tokenized, "extracting embeddings for tokens")]

        bert_embeddings = pd.DataFrame(node_text_tensors, index=graph.get_node_names())

    if not bert_embeddings.empty:
        node_embedding = pd.concat([node_embedding, bert_embeddings],
                                   axis=1,
                                   ignore_index=False)

    node_embedding.to_csv(embedding_outfile, header=False)
    return None


def merge_and_write_complete_node_data(
        original_nodes_file: str, node_data: pd.DataFrame, outfile: str):
    complete_node_data = get_node_data(original_nodes_file)
    node_data = node_data.merge(complete_node_data, how="left", on="id",
                                suffixes=("", "_y")).drop("category_y", axis=1)
    node_data.to_csv(outfile, sep="\t", index=False)
